// -------------------------------------------------------------------------------- 
//!  @file  md.c                                                                    
//!  メガドライブパッドソースファイル                                               
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include <stdio.h>
#include <string.h>
#include "pico/stdlib.h"
#include "config.h"
#include "pad.h"
#include "md.h"


#ifdef  PAD_MD_CFG_PIN_SEL

	enum {
		SEGATAP_RETRYCOUNT			=	750,						//!< セガタップリトライ回数（約 60us）
		SEGATAP_PAD_MAX				=	4							//!< セガタップ最大パッド接続数
	} ;

	// セガタップ種類
	enum {
		SEGATAP_TYPE_3B				=	0,							//!< ３ボタン
		SEGATAP_TYPE_6B				=	1,							//!< ６ボタン
		SEGATAP_TYPE_UNCONNECTED	=	15							//!< 未接続
	} ;


// -------------------------------------------------------------------------------- 
//!  1usec 待ち                                                                     
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC wait1us( void )
{
	volatile int	index ;


	for(index=0; index<11; index++) {
		asm volatile("nop") ;
	}
}


// -------------------------------------------------------------------------------- 
//!  ポート設定                                                                     
//!                                                                                 
//!  @param [in] gpio   GPIO 番号                                                   
//!  @param [in] value  値                                                          
//!  @return            なし                                                        
// -------------------------------------------------------------------------------- 
static void RAMFUNC setPort(uint gpio, bool value)
{
	gpio_put(gpio, value) ;

	wait1us() ;
}


// -------------------------------------------------------------------------------- 
//!  セガタップのデータを取得                                                       
//!                                                                                 
//!  @param [in] clk  クロック                                                      
//!  @return          データ                                                        
// -------------------------------------------------------------------------------- 
static uint8_t RAMFUNC getSegatapData( uint8_t clk )
{
	int		count ;


	setPort(PAD_MD_CFG_PIN_D5, clk) ;

	count = SEGATAP_RETRYCOUNT ;

	do {
		if( gpio_get( PAD_MD_CFG_PIN_D4 ) == clk ) {
			break ;
		}
	} while( --count ) ;

	return (uint8_t )((gpio_get_all()>>PAD_MD_CFG_PIN_D0)&0x0f) ;
}


// -------------------------------------------------------------------------------- 
//!  セガタップのパッド情報を取得                                                   
//!                                                                                 
//!  @param [out]  p_padInfo  パッド情報格納アドレス                                
//!  @return       なし                                                             
// -------------------------------------------------------------------------------- 
static inline void getSegatapPadInfo( PADINFO *p_padInfo )
{
	uint8_t		type, data[6], status[2] ;
	int			index, clk ;


	gpio_put(PAD_MD_CFG_PIN_D5,     1) ;
	gpio_set_dir(PAD_MD_CFG_PIN_D5, GPIO_OUT) ;
	setPort(PAD_MD_CFG_PIN_SEL,     1) ;
	setPort(PAD_MD_CFG_PIN_SEL,     0) ;

	// 固定値＋ID 取得
	for(index=0; index<6; index++) {
		data[index] = getSegatapData( index & 1 ) ;
	}

	// 固定値異常なら未接続
	if( data[0] | data[1] ) {
		memset(&data[2], SEGATAP_TYPE_UNCONNECTED, SEGATAP_PAD_MAX) ;
	}

	// パッド状態取得
	clk = 0 ;
	for(index=0; index<SEGATAP_PAD_MAX; index++,p_padInfo++) {
		type = data[2+index] ;
		if( type > SEGATAP_TYPE_6B ) {

			// 未接続
			p_padInfo->type   = PAD_TYPE_UNKNOWN ;
			p_padInfo->status = 0 ;
			continue ;
		}

		status[0] = getSegatapData( clk ) ^ 0x0f ;
		clk      ^= 1 ;

		status[1] = getSegatapData( clk ) ^ 0x0f ;
		clk      ^= 1 ;

		p_padInfo->type   = PAD_TYPE_MD ;
		p_padInfo->status = ((uint16_t )((status[0] >> 2) & 1) <<  PAD_LEFT_)   |
		                    ((uint16_t )((status[0] >> 3) & 1) <<  PAD_RIGHT_)  |
		                    ((uint16_t )((status[0] >> 0) & 1) <<  PAD_UP_)     |
		                    ((uint16_t )((status[0] >> 1) & 1) <<  PAD_DOWN_)   |
		                    ((uint16_t )((status[1] >> 3) & 1) <<  PAD_START_)  |
		                    ((uint16_t )((status[1] >> 2) & 1) << (PAD_BTN_+0)) |
		                    ((uint16_t )((status[1] >> 0) & 1) << (PAD_BTN_+1)) |
		                    ((uint16_t )((status[1] >> 1) & 1) << (PAD_BTN_+2)) ;

		if( type == SEGATAP_TYPE_6B ) {
			status[0] = getSegatapData( clk ) ^ 0x0f ;
			clk      ^= 1 ;

			p_padInfo->type    = PAD_TYPE_MD_6B ;
			p_padInfo->status |= ((uint16_t )((status[0] >> 3) & 1) <<  PAD_SELECT_) |
			                     ((uint16_t )((status[0] >> 2) & 1) << (PAD_BTN_+3)) |
			                     ((uint16_t )((status[0] >> 1) & 1) << (PAD_BTN_+4)) |
			                     ((uint16_t )((status[0] >> 0) & 1) << (PAD_BTN_+5)) ;
		}
	}

	setPort(PAD_MD_CFG_PIN_SEL,     1) ;
	gpio_set_dir(PAD_MD_CFG_PIN_D5, GPIO_IN) ;
}


// -------------------------------------------------------------------------------- 
//!  初期化                                                                         
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void pad_md_init( void )
{
	gpio_init( PAD_MD_CFG_PIN_SEL ) ;
	gpio_init( PAD_MD_CFG_PIN_D0 ) ;
	gpio_init( PAD_MD_CFG_PIN_D1 ) ;
	gpio_init( PAD_MD_CFG_PIN_D2 ) ;
	gpio_init( PAD_MD_CFG_PIN_D3 ) ;
	gpio_init( PAD_MD_CFG_PIN_D4 ) ;
	gpio_init( PAD_MD_CFG_PIN_D5 ) ;

	gpio_disable_pulls( PAD_MD_CFG_PIN_SEL ) ;
	gpio_pull_up( PAD_MD_CFG_PIN_D0 ) ;
	gpio_pull_up( PAD_MD_CFG_PIN_D1 ) ;
	gpio_pull_up( PAD_MD_CFG_PIN_D2 ) ;
	gpio_pull_up( PAD_MD_CFG_PIN_D3 ) ;
	gpio_pull_up( PAD_MD_CFG_PIN_D4 ) ;
	gpio_pull_up( PAD_MD_CFG_PIN_D5 ) ;

	gpio_put(PAD_MD_CFG_PIN_SEL, 1) ;

	gpio_set_dir(PAD_MD_CFG_PIN_SEL, GPIO_OUT) ;
	gpio_set_dir(PAD_MD_CFG_PIN_D0,  GPIO_IN) ;
	gpio_set_dir(PAD_MD_CFG_PIN_D1,  GPIO_IN) ;
	gpio_set_dir(PAD_MD_CFG_PIN_D2,  GPIO_IN) ;
	gpio_set_dir(PAD_MD_CFG_PIN_D3,  GPIO_IN) ;
	gpio_set_dir(PAD_MD_CFG_PIN_D4,  GPIO_IN) ;
	gpio_set_dir(PAD_MD_CFG_PIN_D5,  GPIO_IN) ;
}


// -------------------------------------------------------------------------------- 
//!  パッド状態を取得                                                               
//!                                                                                 
//!  @param [in,out] p_padInfo  パッド情報                                          
//!  @return                    なし                                                
// -------------------------------------------------------------------------------- 
void RAMFUNC pad_md_get( PADINFO *p_padInfo )
{
	uint8_t		status[4] ;

	// ス | Ａ | ０ | ０ | 下 | 上
	setPort(PAD_MD_CFG_PIN_SEL, 0) ;
	status[0] = (uint8_t )((~gpio_get_all() >> PAD_MD_CFG_PIN_D0) & 0x3f) ;

	// Ｃ | Ｂ | 右 | 左 | 下 | 上
	setPort(PAD_MD_CFG_PIN_SEL, 1) ;
	status[1] = (uint8_t )((~gpio_get_all() >> PAD_MD_CFG_PIN_D0) & 0x3f) ;

	switch( (status[0]<<8) | status[1] ) {
		case 0x0000:	// 未接続（0x3f3f）
		{
			p_padInfo->type   = PAD_TYPE_UNKNOWN ;
			p_padInfo->status = 0 ;
			break ;
		}

		case 0x000c:	// セガタップ（0x3f33）
		{
			PADINFO		padInfo[SEGATAP_PAD_MAX] ;


			getSegatapPadInfo( &padInfo[0] ) ;

			*p_padInfo = padInfo[0] ;
			break ;
		}

		default:
		{
			// ス | Ａ | ０ | ０ | 下 | 上
			setPort(PAD_MD_CFG_PIN_SEL, 0) ;

			// Ｃ | Ｂ | 右 | 左 | 下 | 上
			setPort(PAD_MD_CFG_PIN_SEL, 1) ;

			// ス | Ａ | ０ | ０ | ０ | ０
			setPort(PAD_MD_CFG_PIN_SEL, 0) ;
			status[2] = (uint8_t )((~gpio_get_all() >> PAD_MD_CFG_PIN_D0) & 0x3f) ;

			// １ | １ | モ | Ｘ | Ｙ | Ｚ
			setPort(PAD_MD_CFG_PIN_SEL, 1) ;
			status[3] = (uint8_t )((~gpio_get_all() >> PAD_MD_CFG_PIN_D0) & 0x3f) ;

			p_padInfo->type   = PAD_TYPE_MD ;
			p_padInfo->status = ((uint16_t )((status[1] >> 2) & 1) <<  PAD_LEFT_)   |
			                    ((uint16_t )((status[1] >> 3) & 1) <<  PAD_RIGHT_)  |
			                    ((uint16_t )((status[1] >> 0) & 1) <<  PAD_UP_)     |
			                    ((uint16_t )((status[1] >> 1) & 1) <<  PAD_DOWN_)   |
			                    ((uint16_t )((status[0] >> 5) & 1) <<  PAD_START_)  |
			                    ((uint16_t )((status[0] >> 4) & 1) << (PAD_BTN_+0)) |
			                    ((uint16_t )((status[1] >> 4) & 1) << (PAD_BTN_+1)) |
			                    ((uint16_t )((status[1] >> 5) & 1) << (PAD_BTN_+2)) ;

			// 6B パッド
			if( (status[2]&0x0f) == 0x0f ) {
				p_padInfo->type    = PAD_TYPE_MD_6B ;
				p_padInfo->status |= ((uint16_t )((status[3] >> 3) & 1) <<  PAD_SELECT_) |
				                     ((uint16_t )((status[3] >> 2) & 1) << (PAD_BTN_+3)) |
				                     ((uint16_t )((status[3] >> 1) & 1) << (PAD_BTN_+4)) |
				                     ((uint16_t )((status[3] >> 0) & 1) << (PAD_BTN_+5)) ;
			}
			break ;
		}
	}
}

#endif//PAD_MD_CFG_PIN_SEL
