// -------------------------------------------------------------------------------- 
//!  @file  pad.c                                                                   
//!  パッドソースファイル                                                           
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include <stdio.h>
#include <string.h>
#include "pico/stdlib.h"
#include "hardware/adc.h"
#include "config.h"
#include "pad.h"
#include "serial.h"
#include "pce.h"
#include "md.h"
#include "sfc.h"
#include "ps.h"
#include "classic.h"

	//! パッド処理
	typedef struct {
		void (*init)( void  ) ;										//!< 初期化処理
		void (*get)( PADINFO *p_padInfo ) ;							//!< パッド状態取得処理
	} PADPROC ;


PADINFO								g_padInfo ;						//!< パッド情報
static PADPROC						m_padProc ;						//!< パッド処理


// -------------------------------------------------------------------------------- 
//!  パッド ID を取得                                                               
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static uint8_t getPadId( void )
{
	int		retry, index, count, id ;


	adc_init() ;
	adc_gpio_init( 28 ) ;
	adc_select_input( 2 ) ;

	id    = PAD_ID_UNUSED ;
	count = 0 ;

	for(retry=0; retry<200; retry++) {
		sleep_ms( 1 ) ;

		index = pad_getId() ;
		if( id == index ) {
			if( ++count == 10 ) {
				return id ;
			}
		} else {
			id    = index ;
			count = 0 ;
		}
	}

	return PAD_ID_UNUSED ;
}


// -------------------------------------------------------------------------------- 
//!  初期化                                                                         
//!                                                                                 
//!  @return パッド ID                                                              
// -------------------------------------------------------------------------------- 
uint8_t pad_init( void )
{
	static const PADPROC	padProc[]={
#ifdef  PAD_SERIAL_CFG_BAUDRATE
								{ NULL,             pad_serial_get  },	// PAD_ID_SERIAL
#else
								{ NULL,             NULL            },
#endif//PAD_SERIAL_CFG_BAUDRATE

#ifdef  PAD_PCE_CFG_PIN_SEL
								{ pad_pce_init,     pad_pce_get     },	// PAD_ID_PCE
#else
								{ NULL,             NULL            },
#endif//PAD_PCE_CFG_PIN_SEL

#ifdef  PAD_MD_CFG_PIN_SEL
								{ pad_md_init,      pad_md_get      },	// PAD_ID_MD
#else
								{ NULL,             NULL            },
#endif//PAD_MD_CFG_PIN_SEL

#ifdef  PAD_SFC_CFG_PIN_CLK
								{ pad_sfc_init,     pad_sfc_get     },	// PAD_ID_SFC
#else
								{ NULL,             NULL            },
#endif//PAD_SFC_CFG_PIN_CLK

#ifdef  PAD_PS_CFG_PIN_ACK
								{ pad_ps_init,      pad_ps_get      },	// PAD_ID_PS
#else
								{ NULL,             NULL            },
#endif//PAD_PS_CFG_PIN_ACK

#ifdef  PAD_CLASSIC_CFG_PIN_DETECT
								{ pad_classic_init, pad_classic_get },	// PAD_ID_CLASSIC
#else
								{ NULL,             NULL            },
#endif//PAD_CLASSIC_CFG_PIN_DETECT

								{ NULL,             NULL            },	// 未定義
								{ NULL,             NULL            },	// 未定義
							} ;
	uint8_t					padId ;


	memset(&g_padInfo, 0, sizeof( g_padInfo )) ;
	g_padInfo.analog[0] = 0x80 ;
	g_padInfo.analog[1] = 0x80 ;
	g_padInfo.analog[2] = 0x80 ;
	g_padInfo.analog[3] = 0x80 ;

#if     (PAD_CFG_ID==-1)
	padId = getPadId() ;
#else
	padId = PAD_CFG_ID ;
#endif//(PAD_CFG_ID==-1)

	m_padProc = padProc[padId] ;

#ifdef  PAD_SERIAL_CFG_BAUDRATE
	if( padId == PAD_ID_SERIAL ) {
		pad_serial_init() ;
	}
#endif//PAD_SERIAL_CFG_BAUDRATE

	return padId ;
}


// -------------------------------------------------------------------------------- 
//!  パッド読み取り開始                                                             
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void pad_start( void )
{
	if( m_padProc.init != NULL ) {
		m_padProc.init() ;
	}
}


// -------------------------------------------------------------------------------- 
//!  パッド状態を更新                                                               
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void RAMFUNC pad_update( void )
{
	if( m_padProc.get != NULL ) {
		m_padProc.get( &g_padInfo ) ;
	}
}


// -------------------------------------------------------------------------------- 
//!  パッド ID を取得                                                               
//!                                                                                 
//!  @return パッド ID                                                              
// -------------------------------------------------------------------------------- 
uint8_t RAMFUNC pad_getId( void )
{
#if     (PAD_CFG_ID==-1)
	static const uint16_t	threshold[]={
								800,		// PAD_ID_SERIAL
								1300,		// PAD_ID_PCE
								1600,		// PAD_ID_MD
								1900,		// PAD_ID_SFC
								2100,		// PAD_ID_PS
								2400,		// PAD_ID_CLASSIC
								2580		// 未定義
							} ;
	uint16_t				value ;
	int						index ;


	value = adc_read() ;
	for(index=0; index<sizeof( threshold )/sizeof( threshold[0] ); index++) {
		if( value < threshold[index] ) {
			return (uint8_t )index ;
		}
	}

	return PAD_ID_UNUSED ;
#else
	return PAD_CFG_ID ;
#endif//(PAD_CFG_ID==-1)
}
