// -------------------------------------------------------------------------------- 
//!  @file  ps.c                                                                    
//!  PS コントロールパッドソースファイル                                            
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include <stdio.h>
#include <string.h>
#include "pico/stdlib.h"
#include "hardware/spi.h"
#include "config.h"
#include "pad.h"
#include "ps.h"


#ifdef  PAD_PS_CFG_PIN_ACK

	enum {
		SPI_CLK						=	250 * 1000,					//!< SPI クロック
		SPI_WAIT_XFER				=	125							//!< SPI 送受信後の待ちカウンタ（約16us）
	} ;

	// パッド種類
	enum {
		TYPE_DIGITAL				=	0x41,						//!< デジタル
		TYPE_ANALOG_R				=	0x73,						//!< アナログ赤
		TYPE_ANALOG_G				=	0x53,						//!< アナログ緑
		TYPE_NEGCON					=	0x23,						//!< ネジコン
		TYPE_MOUSE					=	0x12,						//!< マウス
		TYPE_MAX
	} ;


// -------------------------------------------------------------------------------- 
//!  1usec 待ち                                                                     
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC wait10us( void )
{
	volatile int	index ;


	for(index=0; index<110; index++) {
		asm volatile("nop") ;
	}
}


// -------------------------------------------------------------------------------- 
//!  ビットスワップ                                                                 
//!                                                                                 
//!  @param [in] value  値                                                          
//!  @return            ビットスワップ後の値                                        
// -------------------------------------------------------------------------------- 
static inline uint8_t swapBit8( uint8_t value )
{
	value = ((value & 0x55) << 1) | ((value & 0xaa) >> 1) ;
	value = ((value & 0x33) << 2) | ((value & 0xcc) >> 2) ;
	value = ( value         << 4) | ( value         >> 4) ;

	return value ;
}


// -------------------------------------------------------------------------------- 
//!  １バイト転送                                                                   
//!                                                                                 
//!  @param [in] send  送信データ                                                   
//!  @param [in] ack   ACK 待ち                                                     
//!  @return           受信データ                                                   
// -------------------------------------------------------------------------------- 
static uint8_t RAMFUNC xferByte(uint8_t send, bool ack)
{
	uint8_t			recv ;
	volatile int	index ;


	spi_read_blocking(PAD_PS_CFG_SPI, send, &recv, 1) ;

	recv = swapBit8( recv ) ;

	// ACK 待ち
	if( ack != false ) {
		for(index=0; index<SPI_WAIT_XFER; index++) {
			if( !gpio_get( PAD_PS_CFG_PIN_ACK ) ) {
				break ;
			}
		}
	}

	return recv ;
}


// -------------------------------------------------------------------------------- 
//!  初期化                                                                         
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void pad_ps_init( void )
{
	gpio_init( PAD_PS_CFG_PIN_ACK ) ;
	gpio_init( PAD_PS_CFG_PIN_SS ) ;
	gpio_set_function(PAD_PS_CFG_PIN_MISO, GPIO_FUNC_SPI) ;
	gpio_set_function(PAD_PS_CFG_PIN_SCLK, GPIO_FUNC_SPI) ;
	gpio_set_function(PAD_PS_CFG_PIN_MOSI, GPIO_FUNC_SPI) ;

	gpio_disable_pulls( PAD_PS_CFG_PIN_ACK ) ;
	gpio_disable_pulls( PAD_PS_CFG_PIN_MISO ) ;
	gpio_disable_pulls( PAD_PS_CFG_PIN_SS ) ;
	gpio_disable_pulls( PAD_PS_CFG_PIN_SCLK ) ;
	gpio_disable_pulls( PAD_PS_CFG_PIN_MOSI ) ;

	gpio_put(PAD_PS_CFG_PIN_SS, 1) ;
	gpio_set_dir(PAD_PS_CFG_PIN_SS, GPIO_OUT) ;

	spi_init(PAD_PS_CFG_SPI, SPI_CLK) ;
	spi_set_format(PAD_PS_CFG_SPI, 8, SPI_CPOL_1, SPI_CPHA_1, SPI_MSB_FIRST) ;
}


// -------------------------------------------------------------------------------- 
//!  パッド状態を取得                                                               
//!                                                                                 
//!  @param [in,out] p_padInfo  パッド情報                                          
//!  @return                    なし                                                
// -------------------------------------------------------------------------------- 
void RAMFUNC pad_ps_get( PADINFO *p_padInfo )
{
	uint8_t		status[5] ;


	gpio_put(PAD_PS_CFG_PIN_SS, 0) ;

	wait10us() ;

	status[0] = xferByte(0x80, true) ;
	status[1] = xferByte(0x42, true) ;
	status[2] = xferByte(0x00, true) ;

	switch( ((uint32_t )status[1]<<8) | status[2] ) {
		case (TYPE_DIGITAL <<8)|0x5a:
		case (TYPE_ANALOG_R<<8)|0x5a:
		case (TYPE_ANALOG_G<<8)|0x5a:
		{
			uint8_t		analog[4] ;


			status[3] = ~xferByte(0x00, true) ;		// 左 | 下 | 右 | 上 | ス | Ｒ | Ｌ | セ
			status[4] = ~xferByte(0x00, true) ;		// □ | × | 〇 | △ | R1 | L1 | R2 | L2

			p_padInfo->status = ((uint16_t )((status[3] >> 7) & 1) <<  PAD_LEFT_)    |
			                    ((uint16_t )((status[3] >> 5) & 1) <<  PAD_RIGHT_)   |
			                    ((uint16_t )((status[3] >> 4) & 1) <<  PAD_UP_)      |
			                    ((uint16_t )((status[3] >> 6) & 1) <<  PAD_DOWN_)    |
			                    ((uint16_t )((status[3] >> 0) & 1) <<  PAD_SELECT_)  |
			                    ((uint16_t )((status[3] >> 3) & 1) <<  PAD_START_)   |
			                    ((uint16_t )((status[4] >> 6) & 1) << (PAD_BTN_+ 0)) |
			                    ((uint16_t )((status[4] >> 5) & 1) << (PAD_BTN_+ 1)) |
			                    ((uint16_t )((status[4] >> 2) & 1) << (PAD_BTN_+ 2)) |
			                    ((uint16_t )((status[4] >> 7) & 1) << (PAD_BTN_+ 3)) |
			                    ((uint16_t )((status[4] >> 4) & 1) << (PAD_BTN_+ 4)) |
			                    ((uint16_t )((status[4] >> 3) & 1) << (PAD_BTN_+ 5)) |
			                    ((uint16_t )((status[3] >> 1) & 1) << (PAD_BTN_+ 6)) |
			                    ((uint16_t )((status[3] >> 2) & 1) << (PAD_BTN_+ 7)) |
			                    ((uint16_t )((status[4] >> 0) & 1) <<  PAD_LB_)      |
			                    ((uint16_t )((status[4] >> 1) & 1) <<  PAD_RB_) ;

			if( status[1] == TYPE_DIGITAL ) {

				p_padInfo->type      = PAD_TYPE_PS ;
				p_padInfo->analog[0] = 0x80 ;
				p_padInfo->analog[1] = 0x80 ;
				p_padInfo->analog[2] = 0x80 ;
				p_padInfo->analog[3] = 0x80 ;

			} else {

				p_padInfo->type = PAD_TYPE_PS_ANALOG ;
				analog[0]       = xferByte(0x00, true) ;		// 右アナログ水平方向
				analog[1]       = xferByte(0x00, true) ;		// 右アナログ垂直方向
				analog[2]       = xferByte(0x00, true) ;		// 左アナログ水平方向
				analog[3]       = xferByte(0x00, false) ;		// 左アナログ垂直方向

				// アナログの取得は VBlank 後になるため転送完了後にまとめてコピー（VSync 内でパッド情報を取得している場合アナログは１フレーム遅れる）
				p_padInfo->analog[0] = analog[2] ;
				p_padInfo->analog[1] = analog[3] ;
				p_padInfo->analog[2] = analog[0] ;
				p_padInfo->analog[3] = analog[1] ;

			}
			break ;
		}

		default:
		{
			p_padInfo->type      = PAD_TYPE_UNKNOWN ;
			p_padInfo->status    = 0 ;
			p_padInfo->analog[0] = 0x80 ;
			p_padInfo->analog[1] = 0x80 ;
			p_padInfo->analog[2] = 0x80 ;
			p_padInfo->analog[3] = 0x80 ;
			break ;
		}
	}

	gpio_put(PAD_PS_CFG_PIN_SS, 1) ;
}

#endif//PAD_PS_CFG_PIN_ACK
