// -------------------------------------------------------------------------------- 
//!  @file  serial.c                                                                
//!  シリアル通信パッドソースファイル                                               
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include <stdio.h>
#include <string.h>
#include "pico/stdlib.h"
#include "config.h"
#include "pad.h"
#include "hardware/irq.h"


#ifdef  PAD_SERIAL_CFG_BAUDRATE

	enum {
		PADDATA_SIZE				=	8,							//!< パッドデータサイズ
		PADDATA_BEGIN				=	0xf8,						//!< パッドデータ開始
		PADDATA_END					=	0xf9						//!< パッドデータ終了
	} ;


static uint8_t						m_start ;						//!< 読み取り開始フラグ
static uint8_t						m_recvDataCount ;				//!< 受信データ数
static uint8_t						m_recvData[PADDATA_SIZE] ;		//!< 受信データ
static PADINFO						m_padInfo ;						//!< パッド情報


// -------------------------------------------------------------------------------- 
//!  UART 割り込みサービスルーチン                                                  
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC isr_uart_rx( void )
{
	uint8_t		data ;


	while( uart_is_readable( uart0 ) ) {
		data = uart_get_hw( uart0 )->dr ;

		// 未受信
		if( !m_start ) {
			m_start         = (data == PADDATA_BEGIN) ;
			m_recvDataCount = 0 ;
			continue ;
		}

		// データ受信
		if( m_recvDataCount < PADDATA_SIZE ) {
			m_recvData[m_recvDataCount++] = data ;
			continue ;
		}

		// 受信完了
		if( data == PADDATA_END ) {
			m_padInfo.type      = PAD_TYPE_SERIAL ;
			m_padInfo.status    = ((uint16_t )m_recvData[1] << 8) | m_recvData[0] ;
			m_padInfo.analog[0] = m_recvData[2] ;
			m_padInfo.analog[1] = m_recvData[3] ;
			m_padInfo.analog[2] = m_recvData[4] ;
			m_padInfo.analog[3] = m_recvData[5] ;

		} else {
			m_padInfo.type      = PAD_TYPE_UNKNOWN ;
			m_padInfo.status    = 0 ;
			m_padInfo.analog[0] = 0x80 ;
			m_padInfo.analog[1] = 0x80 ;
			m_padInfo.analog[2] = 0x80 ;
			m_padInfo.analog[3] = 0x80 ;
		}
		m_start = 0 ;
	}
}


// -------------------------------------------------------------------------------- 
//!  初期化                                                                         
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void pad_serial_init( void )
{
	memset(&m_padInfo, 0, sizeof( m_padInfo )) ;
	m_start = 0 ;

	uart_init(PAD_SERIAL_CFG_UART, PAD_SERIAL_CFG_BAUDRATE) ;

	gpio_set_function(PAD_SERIAL_CFG_PIN_TX, GPIO_FUNC_UART) ;
	gpio_set_function(PAD_SERIAL_CFG_PIN_RX, GPIO_FUNC_UART) ;

	uart_set_hw_flow(PAD_SERIAL_CFG_UART, false, false) ;
	uart_set_format(PAD_SERIAL_CFG_UART, 8, 1, UART_PARITY_NONE) ;
	uart_set_fifo_enabled(PAD_SERIAL_CFG_UART, false) ;

	irq_set_exclusive_handler(PAD_SERIAL_CFG_UART_IRQ, isr_uart_rx) ;
	irq_set_enabled(PAD_SERIAL_CFG_UART_IRQ, true) ;

	uart_set_irq_enables(PAD_SERIAL_CFG_UART, true, false) ;
}


// -------------------------------------------------------------------------------- 
//!  パッド状態を取得                                                               
//!                                                                                 
//!  @param [in,out] p_padInfo  パッド情報                                          
//!  @return                    なし                                                
//!  @attention 割り込み禁止状態で呼ぶこと。                                        
// -------------------------------------------------------------------------------- 
void RAMFUNC pad_serial_get( PADINFO *p_padInfo )
{
	// ※本当はコピーにスピンロックを使用した方が安全

	*p_padInfo = m_padInfo ;
}

#endif//PAD_SERIAL_CFG_BAUDRATE
