// -------------------------------------------------------------------------------- 
//!  @file  ay-3-8910.c                                                             
//!  AY-3-8910 音源 PCM データジェネレータソースファイル                            
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include <stdio.h>
#include <string.h>
#include "pico/stdlib.h"
#include "config.h"
#include "ay-3-8910.h"


#ifdef  PCMAY38910_CFG_BASECLK

	enum {
		CLK							=	(PCMAY38910_CFG_BASECLK * PCMAY38910_CFG_TIME / 16 + (1000000 / 2)) / 1000000
																	//!< クロック
	} ;

	//! Envelope Shape Cycle
	enum {
		ENVELOPE_REG_CONTINUE_		=	3,							//!< Continue
		ENVELOPE_REG_ATTACK_		=	2,							//!< Attack
		ENVELOPE_REG_ALTERNATE_		=	1,							//!< Alternate
		ENVELOPE_REG_HOLD_			=	0,							//!< Hold

		ENVELOPE_REG_CONTINUE		=	1 << ENVELOPE_REG_CONTINUE_,
		ENVELOPE_REG_ATTACK			=	1 << ENVELOPE_REG_ATTACK_,
		ENVELOPE_REG_ALTERNATE		=	1 << ENVELOPE_REG_ALTERNATE_,
		ENVELOPE_REG_HOLD			=	1 << ENVELOPE_REG_HOLD_
	} ;


static const uint16_t				m_volume[16]={					//!< 音量変換データ
										     0,
										   8/4,
										  11/4,
										  16/4,
										  22/4,
										  32/4,
										  45/4,
										  64/4,
										  90/4,
										 128/4,
										 181/4,
										 256/4,
										 362/4,
										 512/4,
										 724/4,
										1024/4
									} ;


// -------------------------------------------------------------------------------- 
//!  リセット                                                                       
//!                                                                                 
//!  @param [out] p  AY-3-8910 情報                                                 
//!  @return         なし                                                           
// -------------------------------------------------------------------------------- 
void pcm_ay_3_8910_reset( PCMAY38910INFO *p )
{
	memset(p, 0, sizeof( PCMAY38910INFO )) ;
	p->channelInfo[0].reg.n32 = 0x09000000 ;
	p->channelInfo[1].reg.n32 = 0x09000000 ;
	p->channelInfo[2].reg.n32 = 0x09000000 ;
}


// -------------------------------------------------------------------------------- 
//!  更新                                                                           
//!                                                                                 
//!  @param [in,out] p  AY-3-8910 情報                                              
//!  @return            出力データ                                                  
// -------------------------------------------------------------------------------- 
int16_t RAMFUNC pcm_ay_3_8910_update( PCMAY38910INFO *p )
{
	PCMAY38910REG			reg ;
	PCMAY38910CHANNELINFO	*p_channelInfo, *p_channelInfoEnd ;
	uint8_t					ctrl, volume ;
	int32_t					output ;
	uint32_t				freq ;


	reg = p->reg ;

	// ノイズ
	freq = reg.n8[3] ;
	if( freq ) {
		p->noise.freqCount += CLK ;
		while( p->noise.freqCount >= freq ) {
			p->noise.freqCount -= freq ;
			p->noise.output    = ((p->noise.output & 0xffff) << 1) | (((p->noise.output >> 16) ^ (p->noise.output >> 13) ^ 1) & 1) ;
		}
	}

	// エンベロープ
	if( p->envelope.update ) {
		p->envelope.update = 0 ;
		p->envelope.reg    = reg.n8[2] ;

		ctrl = reg.n8[2] ;
		if( !(reg.n8[2]&ENVELOPE_REG_CONTINUE) ) {
			ctrl = ENVELOPE_REG_CONTINUE | ENVELOPE_REG_HOLD ;
			if( reg.n8[2] & ENVELOPE_REG_CONTINUE ) {
				ctrl = ENVELOPE_REG_CONTINUE | ENVELOPE_REG_ATTACK | ENVELOPE_REG_ALTERNATE | ENVELOPE_REG_HOLD ;
			}
		}

		p->envelope.ctrl   = ctrl & ~ENVELOPE_REG_HOLD ;
		p->envelope.volume = ((~ctrl >> ENVELOPE_REG_ATTACK_) & 1) * 15 ;
	}

	freq = reg.n16[0] ;
	if( freq > 1 ) {
		p->envelope.freqCount += CLK ;
		while( p->envelope.freqCount >= freq ) {
			p->envelope.freqCount -= freq ;

			if( p->envelope.ctrl & ENVELOPE_REG_HOLD ) {
				break ;
			}

			if( p->envelope.ctrl & ENVELOPE_REG_ATTACK ) {
				p->envelope.volume++ ;
			} else {
				p->envelope.volume-- ;
			}

			if( !(p->envelope.volume&~15) ) {
				continue ;
			}

			if( p->envelope.ctrl & ENVELOPE_REG_ALTERNATE ) {
				p->envelope.ctrl ^= ENVELOPE_REG_ATTACK ;
			}

			if( p->envelope.reg & ENVELOPE_REG_HOLD ) {
				p->envelope.ctrl   = ENVELOPE_REG_HOLD ;
				p->envelope.volume = (( p->envelope.ctrl >> ENVELOPE_REG_ATTACK_) & 1) * 15 ;
			} else {
				p->envelope.volume = ((~p->envelope.ctrl >> ENVELOPE_REG_ATTACK_) & 1) * 15 ;
			}
		}
	}

	// チャンネル
	p_channelInfo    = &p->channelInfo[0] ;
	p_channelInfoEnd = &p->channelInfo[PCMAY38910_CHANNEL_COUNT] ;
	output           = 0 ;

	do {
		reg  = p_channelInfo->reg ;
		freq = reg.n16[0] ;

		if( freq > 1 ) {
			p_channelInfo->freqCount += CLK ;
			while( p_channelInfo->freqCount >= freq ) {
				p_channelInfo->freqCount -= freq ;
				p_channelInfo->output    ^= 1 ;
			}
		}

		do {
			// ノイズ
			if( !(reg.n16[1]&0x0800) ) {
				if( p->noise.output & (1<<16) ) {
					break ;
				}
			}

			if( reg.n16[1] & 0x0100 ) {
				break ;
			}

			if( reg.n16[1] & 0x0010 ) {		// エンベロープ
				volume = p->envelope.volume ;
			} else {						// トーン
				volume = reg.n16[1] & 15 ;
			}

			if( p_channelInfo->output ) {
				output -= m_volume[volume] ;
			} else {
				output += m_volume[volume] ;
			}
		} while( 0 ) ;
	} while( ++p_channelInfo != p_channelInfoEnd ) ;

	return (int16_t )output ;
}

#endif//PCMAY38910_CFG_BASECLK
