// -------------------------------------------------------------------------------- 
//!  @file  scc.h                                                                   
//!  SCC 音源 PCM データジェネレータヘッダファイル                                  
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#ifndef __PCM__SCC_H__INCLUDED__
#define __PCM__SCC_H__INCLUDED__


// レジスタ |VVVV|FFFF|FFFF|FFFF|
//        V  音量
//        F  周波数


	enum {
		PCMSCC_WAVE_SIZE			=	128,						//!< 波形サイズ
		PCMSCC_CHANNEL_COUNT		=	5,							//!< チャンネル数
		PCMSCC_CHANNEL_WAVE_SIZE	=	32,							//!< チャンネルあたりの波形サイズ
																	//!< チャンネルあたりの波形サイズマスク
		PCMSCC_CHANNEL_WAVE_MASK	=	PCMSCC_CHANNEL_WAVE_SIZE - 1
	} ;

	//! チャンネル情報
	typedef struct {
		uint16_t					reg ;							//!< レジスタ値
		uint8_t						offset ;						//!< 波形 RAM 開始オフセット
		uint8_t						index ;							//!< 波形 RAM インデックス
		uint16_t					count ;							//!< 周波数カウンタ
	} PCMSCCCHANNELINFO ;

	//! SCC 情報
	typedef struct {
		int8_t						wave[PCMSCC_WAVE_SIZE] ;		//!< 波形 RAM
		PCMSCCCHANNELINFO			channelInfo[PCMSCC_CHANNEL_COUNT] ;
																	//!< チャンネル情報
		uint8_t						enable ;						//!< チャンネル有効フラグ
	} PCMSCCINFO ;


extern void            pcm_scc_reset( PCMSCCINFO *p ) ;
extern int16_t RAMFUNC pcm_scc_update( PCMSCCINFO *p ) ;


// -------------------------------------------------------------------------------- 
//!  チャンネル有効・無効                                                           
//!                                                                                 
//!  @param [out] p       SCC 情報                                                  
//!  @param [in]  enable  チャンネル有効ビット                                      
//!  @return              なし                                                      
// -------------------------------------------------------------------------------- 
static inline void pcm_scc_enable(PCMSCCINFO *p, uint8_t enable)
{
	p->enable = enable & ((1 << PCMSCC_CHANNEL_COUNT) - 1) ;
}


// -------------------------------------------------------------------------------- 
//!  波形データを波形 RAM へコピー                                                  
//!                                                                                 
//!  @param [out] p       SCC 情報                                                  
//!  @param [in]  offset  開始オフセット                                            
//!  @param [in]  p_wave  波形データアドレス                                        
//!  @param [in]  size    波形データサイズ                                          
//!  @return              なし                                                      
// -------------------------------------------------------------------------------- 
static inline void pcm_scc_copyWaveData(PCMSCCINFO *p, uint8_t offset, int8_t const *p_wave, uint8_t size)
{
	memcpy(&p->wave[offset], &p_wave[0], size) ;
}


// -------------------------------------------------------------------------------- 
//!  レジスタ設定                                                                   
//!                                                                                 
//!  @param [out] p       SCC 情報                                                  
//!  @param [in]  channel チャンネル番号                                            
//!  @param [in]  value   レジスタ値                                                
//!  @return              なし                                                      
// -------------------------------------------------------------------------------- 
static inline void pcm_scc_setReg(PCMSCCINFO *p, uint8_t channel, uint16_t value)
{
	p->channelInfo[channel].reg = value ;
}


// -------------------------------------------------------------------------------- 
//!  音量設定                                                                       
//!                                                                                 
//!  @param [out] p       SCC 情報                                                  
//!  @param [in] channel  チャンネル番号                                            
//!  @param [in] volume   音量                                                      
//!  @return              なし                                                      
// -------------------------------------------------------------------------------- 
static inline void pcm_scc_setVolume(PCMSCCINFO *p, uint8_t channel, uint8_t volume)
{
	p->channelInfo[channel].reg = (p->channelInfo[channel].reg & 0x0fff) | ((uint16_t )volume << 12) ;
}


// -------------------------------------------------------------------------------- 
//!  周波数設定                                                                     
//!                                                                                 
//!  @param [out] p       SCC 情報                                                  
//!  @param [in] channel  チャンネル番号                                            
//!  @param [in] freq     周波数                                                    
//!  @return              なし                                                      
// -------------------------------------------------------------------------------- 
static inline void pcm_scc_setFreq(PCMSCCINFO *p, uint8_t channel, uint16_t freq)
{
	p->channelInfo[channel].reg = (p->channelInfo[channel].reg & 0xf000) | (freq & 0x0fff) ;
}


#endif//__PCM__SCC_H__INCLUDED__
